{% extends "!layout.html" %}
{% block extrahead %}
    <script defer data-domain="ancell.in/capytaine" src="https://plausible.io/js/script.js"></script>
    <link rel="stylesheet" href="_static/custom.css" type="text/css" />
    {{super}}
{% endblock %}
