.proof {
  margin-top: 1em;
  margin-bottom: 1em;
}

/* Titles */
.proof-title {
  background-color: #9fb1b4;
  border: 1px solid #86989b;
  color: white;
  font-size: 120%;
}

/* Content */
.proof-content {
  border: 1px solid #9fb1b4;
  background-color: #F5F7F9;
  padding: 0em 1em;
}


/* Toggle proof */
.proof-type-proof .proof-title {
    display: block;
    clear: both;
}

.proof-type-proof .proof-title:after {
    content: " ▼";
}

.proof-type-proof .proof-title.open:after {
    content: " ▲";
}
