# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

from capytaine.meshes.predefined.cylinders import mesh_disk, mesh_horizontal_cylinder, mesh_vertical_cylinder
from capytaine.meshes.predefined.spheres import mesh_sphere
from capytaine.meshes.predefined.rectangles import mesh_rectangle, mesh_parallelepiped
