#!/usr/bin/env python3

__all__ = ["__title__", "__description__", "__version__", "__author__", "__uri__", "__license__"]

__title__ = "capytaine"
__description__ = """Python BEM solver for linear potential flow, based on Nemoh"""

__version__ = "2.1"

__author__ = "Matthieu Ancellin"
__uri__ = "https://github.com/capytaine/capytaine"
__license__ = "GPL-3.0"


if __name__ == "__main__":
    print(__version__)
